# -*- coding: iso-8859-1 -*-
"""
Bayes-Schtzer mit diskretem Prior

Vgl. Skript, Kapitel "Mehr Bayesinaismus"
"""

import random, operator


def faculty(n):
	return reduce(operator.mul, range(1, n+1), 1)

def binomial(up, down):
	return faculty(up)/(faculty(down)*faculty(up-down))

def bernoulli(numSucc, numExp, p):
	return binomial(numExp, numSucc)*p**numSucc*(1-p)**(numExp-numSucc)


class ScaledArray:
	"""Diese Klasse stellt ein "Array" mit nPoints Elementen
	dar, auf das mit Fliekommazahlen aus dem Intervall interval
	zugegriffen werden kann.  Dabei wird das Intervall geteilt
	an nPoints-1 Stellen geteilt und fr jedes der entstehende
	Teilintervalle ein Wert reserviert und zurckgegeben, wenn ein
	Wert fr einen Punkt aus dem betreffenden Intervall angefragt
	wird.

	Vorsicht: Zugriffe "auerhalb" des Array werden stillschweigend
	auf das grte oder kleinste Element gemappt.
	>>> s = ScaledArray(4, (0, 1))
	>>> s[0] = 2
	>>> s[0.51] = s[0.2]+3
	>>> s[0.49] = s[0.65]//2
	>>> s[1] = s[0.3]*2
	>>> print s
	2 2 5 4
	>>> s[1.1] += 3
	>>> print s
	2 2 5 7
	>>> print [v for v in s.iterkeys()]
	[0.125, 0.375, 0.625, 0.875]
	>>> s.normalize()
	>>> print s
	0.125 0.125 0.3125 0.4375
	"""
	def __init__(self, nPoints, interval):
		self.points = [0]*nPoints
		self.interStart, self.interEnd = interval
		self.interWidth = float(interval[1]-interval[0])
		self.subWidth = float(interval[1]-interval[0])/nPoints
		self.nPoints = nPoints
		self.subCenters = [self.interStart+self.subWidth*(0.5+a)
			for a in range(nPoints)]

	def __setitem__(self, index, val):
		self.points[self._unscale(index)] = val
	
	def __getitem__(self, index):
		return self.points[self._unscale(index)]

	def __str__(self):
		return " ".join(map(str, self.points))

	def _unscale(self, floatIndex):
		"""gibt einen ganzzahligen index fr self.points zum floatIndex
		zurck.
		"""
		return int((min(self.interEnd-1e-12, max(self.interStart, floatIndex))
			-self.interStart)/self.interWidth*self.nPoints)

	def iterkeys(self):
		return iter(self.subCenters)
	
	def normalize(self):
		norm = float(sum(self.points))
		self.points = [p/norm for p in self.points]

	def _getArgMaxInt(self):
		maxInd, maxVal = 0, self.points[0]
		for ind, val in enumerate(self.points):
			if val>maxVal:
				maxInd = ind
				maxVal = val
		return maxInd
	

class DiscretePrior(ScaledArray):
	"""Diese Klasse stellt einen diskreten Prior mit nPoints Punkten ber einem
	Intervall interval dar.
	"""
	def __init__(self, nPoints, interval):
		ScaledArray.__init__(self, nPoints, interval)
		for i in self.iterkeys():
			self[i] = 1/float(nPoints)

	def getBeliefInterval(self, level):
		leftInd = rightInd = self._getArgMaxInt()
		totalMass = self.points[leftInd]
		while totalMass<1-level:
			leftVal = leftInd and self.points[leftInd-1]
			rightVal = rightInd<self.nPoints and self.points[rightInd+1]
			if leftVal>rightVal:
				totalMass += leftVal
				leftInd -= 1
			else:
				totalMass += rightVal
				rightInd += 1
		return self.subCenters[leftInd], self.subCenters[rightInd]


class BayesianUpdateBernoulliTester:
	def __init__(self, nPoints, trueProb, log=None):
		self.prior = DiscretePrior(nPoints, (0, 1))
		self.trueProb = trueProb
		self.log = log

	def _logPrior(self):
		if self.log is not None:
			self.log.write(str(self.prior)+"\n")

	def updateFor(self, numberExp, numberSucc):
		for val in self.prior.iterkeys():
			self.prior[val] = bernoulli(
				numberSucc, numberExp, val)*self.prior[val]
		self.prior.normalize()
		self._logPrior()

	def toss(self, numberOfTosses):
		return sum([random.random()<self.trueProb 
			for i in range(numberOfTosses)])

	def getPrior(self):
		return self.prior
	

def runIt(numPoints=20, trueProb=0.5, tossesPerExp=10, totalIters=20,
	level=0.05):
	t = BayesianUpdateBernoulliTester(numPoints, trueProb,
		open("priors.log", "w"))
	for i in range(totalIters):
		t.updateFor(tossesPerExp, t.toss(tossesPerExp))
		try:
			print t.getPrior().getBeliefInterval(level)
		except IndexError:
			print "No confidence interval yet"


def _test():
	import doctest, discPrior
	return doctest.testmod(discPrior)


if __name__=="__main__":
	_test()
	runIt(level=0.05, numPoints=20, totalIters=20, tossesPerExp=10,
		trueProb=0.2)
